import dataclasses
from tqdm import trange
import multiprocessing
import time
import orjson

from michie.object import Object
from michie.worker import Worker, Works
from michie.serialize import serialize, deserialize

class World:
    def __init__(self, *, global_mappers=[], tick_hooks=[], config=None):
        self.config = config
        self.global_mappers = global_mappers
        self.tick_hooks = tick_hooks
        self.global_state = dict(
            tick=0,
            michie=dict(stats=self.init_stats())
        )
        self.objects = []
        self.dict_states = []
        self.state_mappers = dict()
        self.state_mappers_ids = []
        self.transitions = dict()
        self.transitions_ids = []
        self.window = None
        self.render_surface = None
    
    def init_stats(self):
        return dict(
            global_mappers_execution_time=0,
            transitions_submission_time=0,
            transitions_retrieval_time=0,
            transitions_works=0,
            state_mappers_submission_time=0,
            state_mappers_retrieval_time=0,
            state_mappers_works=0,
            render_time=0
        )

    def add_object(self, object):
        assert isinstance(object, Object), "You can only add michie.Object instances"
        object.init["type"] = object.name

        transitions_ids = []
        for transition in object.transitions:
            if not transition.__name__ in self.transitions:
                self.transitions[transition.__name__] = transition
            transitions_ids.append(transition.__name__)
        self.transitions_ids.append(transitions_ids)

        state_mappers_ids = []
        for transition in object.state_mappers:
            if not transition.__name__ in self.state_mappers:
                self.state_mappers[transition.__name__] = transition
            state_mappers_ids.append(transition.__name__)
        self.state_mappers_ids.append(state_mappers_ids)

        self.objects.append(object)
        self.dict_states.append(object.init)

    def transitions_tick(self, *, submit_queue, results_queue):
        assert submit_queue.empty() and results_queue.empty()

        start_submission_time = time.time()
        works = 0
        for id, (state, transitions_ids) in enumerate(zip(self.dict_states, self.transitions_ids)):
            for transition_id in transitions_ids:
                transition = self.transitions[transition_id]
                if transition.requirements(state):
                    work = dict(
                        type = Works.STATE_TRANSITION.value,
                        args = dict(
                            id = id,
                            state = transition.state_map(state),
                            transition_id = transition_id
                        )
                    )
                    try:
                        submit_queue.put(serialize(work))
                    except Exception as e:
                        print("Serialization error for work")
                        print(work)
                        raise e
                    
                    works += 1
        
        end_submission_time = time.time()
        for _ in range(0, works):
            result = results_queue.get()
            result = deserialize(result)
            self.dict_states[result["id"]].update(result["result"])  
        end_retrieve_time = time.time()

        self.global_state["michie"]["stats"]["transitions_submission_time"] = end_submission_time - start_submission_time
        self.global_state["michie"]["stats"]["transitions_retrieval_time"] = end_retrieve_time - end_submission_time
        self.global_state["michie"]["stats"]["transitions_works"] = works

        assert submit_queue.empty() and results_queue.empty()
    
    def map_states(self, *, submit_queue, results_queue):
        assert submit_queue.empty() and results_queue.empty()

        start_submission_time = time.time()
        works = 0
        for id, (state, state_mappers_ids) in enumerate(zip(self.dict_states, self.state_mappers_ids)):
            for state_mapper_id in state_mappers_ids:
                mapper = self.state_mappers[state_mapper_id]
                if mapper.requirements(state):
                    work = dict(
                        type = Works.STATE_MAP.value,
                        args = dict(
                            id = id,
                            state = mapper.state_map(state),
                            global_state = mapper.global_state_map(self.global_state),
                            state_mapper_id = state_mapper_id
                        )
                    )
                    try:
                        submit_queue.put(serialize(work))
                    except Exception as e:
                        print("Serialization error for work")
                        print(work)
                        raise e

                    works += 1
        
        end_submission_time = time.time()
        for _ in range(0, works):
            result = results_queue.get()
            result = deserialize(result)
            self.dict_states[result["id"]].update(result["result"])  
        end_retrieve_time = time.time()

        self.global_state["michie"]["stats"]["state_mappers_submission_time"] = end_submission_time - start_submission_time
        self.global_state["michie"]["stats"]["state_mappers_retrieval_time"] = end_retrieve_time - end_submission_time
        self.global_state["michie"]["stats"]["state_mappers_works"] = works

        assert submit_queue.empty() and results_queue.empty()

    def global_map_states(self):
        start_time = time.time()
        for global_mapper in self.global_mappers:
            self.dict_states = global_mapper.map(self.dict_states, self.global_state) 
        end_time = time.time()
        self.global_state["michie"]["stats"]["global_mappers_execution_time"] = end_time - start_time

    def render(self, *, window, clock, fps=30, background="black"):
        start_time = time.time()
        import pygame
        window.fill(background)
        sprites = [object.sprites for object in self.objects]
        for state, object_sprites in zip(self.dict_states, sprites):
            for object_sprite in object_sprites: object_sprite.draw(window=window, state=state)
        pygame.display.flip()
        if clock is not None: clock.tick(fps)
        end_time = time.time()
        self.global_state["michie"]["stats"]["render_time"] = end_time - start_time

    def run(
            self,
            *,
            workers,
            max_ticks=100,
            render=False,
            render_surface=(800, 600),
            render_fps=None,
            render_background="black"
        ):   
        if render:
            import pygame
            pygame.init()
            window = pygame.display.set_mode(render_surface, pygame.HWSURFACE | pygame.DOUBLEBUF)
            clock = None
            if render_fps is not None:
                clock = pygame.time.Clock()

        submit_queue = multiprocessing.Queue()
        results_queue = multiprocessing.Queue()
        
        workers = [
            Worker(
                id=id,
                submit_queue=submit_queue,
                results_queue=results_queue,
                state_mappers=self.state_mappers,
                transitions=self.transitions
            ) for id in range(0, workers)
        ]
        [worker.start() for worker in workers]
        for hook in self.tick_hooks: hook.start(self.dict_states, self.global_state, window)
        
        for i in trange(0, max_ticks):
            self.global_map_states()
            self.map_states(submit_queue=submit_queue, results_queue=results_queue)
            self.transitions_tick(submit_queue=submit_queue, results_queue=results_queue)
            self.global_state["tick"] += 1

            if render: self.render(
                window=window,
                clock=clock,
                fps=render_fps,
                background=render_background
            )
            for hook in self.tick_hooks: hook.tick(self.dict_states, self.global_state, window)
        
        for hook in self.tick_hooks: hook.end(self.dict_states, self.global_state, window)

        for i in range(0, len(workers)):
            submit_queue.put(serialize(dict(
                type=Works.EXIT.value
            )))


import json

def factory(cls):
    def decorated(**actual_args):
        setattr(cls, "args", actual_args)
        new_class_name = cls.__name__+json.dumps(actual_args)
        return type(new_class_name, cls.__bases__, dict(cls.__dict__))
    return decorated


import numpy as np
import scipy.spatial.distance
import copy

from michie.mappers.globalmapper import GlobalMapper

class NeighboursGlobalMapper(GlobalMapper):
    def __init__(self, *, radius, beacon=None):
        self.radius = radius
        self.beacon = type(self).beacon if beacon is None else beacon
    
    @staticmethod
    def beacon(state):
        beacon = state.copy()
        if "neighbours" in beacon: del beacon["neighbours"]
        return beacon

    def map(self, states, global_state):
        for id, state in enumerate(states):
            dists = state["distances"]
            other_dists = np.concatenate((dists[:id], dists[id+1:]))
            other_states = states[:id] + states[id+1:]

            state["neighbours"] = []
            for id, other_state in enumerate(other_states):
                if other_dists[id] <= self.radius:
                    beacon = self.beacon(other_state)
                    assert "neighbours" not in beacon, "The neighbour beacon must not have the neighbours field"
                    state["neighbours"].append(beacon)

        return states



import numpy as np
import math
from michie.mappers.statemapper import StateMapper

class RangeAndBearingStateMapper(StateMapper):
    @classmethod
    def requirements(cls, state):
        return "neighbours" in state and "position" in state
    
    @classmethod
    def global_state_map(cls, global_state):
        return dict()

    @classmethod
    def state_map(cls, state):
        return dict(
            position=state["position"],
            neighbours=state["neighbours"]
        )
    
    @classmethod
    def beacon(cls, neighbour):
        raise NotImplementedError()

    @classmethod
    def map(cls, id, mapped_state, mapped_global_state):
        range_and_bearing = []
        
        for neighbour in mapped_state["neighbours"]:
            self_x, self_y = mapped_state["position"]["position"]
            self_heading = mapped_state["position"]["heading"]
            neig_x, neig_y = neighbour["position"]["position"]
                        
            point = (neig_x - self_x, neig_y - self_y)
            
            angle = np.arctan2(point[1], point[0])%(2*math.pi) - self_heading
            distance = math.sqrt(point[0]**2 + point[1]**2)
            virtual_point = (distance*np.cos(angle), distance*np.sin(angle))

            range_and_bearing.append(dict(
                distance=float(distance),
                angle=float(angle),
                point=[float(virtual_point[0]), float(virtual_point[1])],
                beacon=cls.beacon(neighbour)
            ))
        
        return dict(
            range_and_bearing=range_and_bearing
        )




class Hook:
    def start(self, states, global_state, window):
        pass

    def tick(self, states, global_state, window):
        pass
    
    def end(self, states, global_state, window):
        pass


import multiprocessing
from enum import Enum

from michie.serialize import serialize, deserialize

class Works(Enum):
    EXIT = 0
    STATE_MAP = 1
    STATE_TRANSITION = 2


class Worker(multiprocessing.Process):
    def __init__(self, *, id, submit_queue, results_queue, state_mappers, transitions):
        super(Worker, self).__init__()
        self.id = id
        self.submit_queue = submit_queue
        self.results_queue = results_queue
        self.state_mappers = state_mappers
        self.transitions = transitions

    def state_map(self, *, id, state, global_state, state_mappers_ids):
        state_mappers = map(lambda t: self.state_mappers[t], state_mappers_ids)
        partial_updates = []

        for state_mapper in state_mappers:
            mapped_state = state_mapper.map(id, state, global_state)
            state.update(mapped_state)
        
        return state

    def run(self):
        while True:
            work = self.submit_queue.get()
            work = deserialize(work)
            
            result = None
            if work["type"] == Works.EXIT.value:
                return
            if work["type"] == Works.STATE_MAP.value:
                result = self.state_mappers[
                    work["args"]["state_mapper_id"]
                ].map(work["args"]["id"], work["args"]["state"], work["args"]["global_state"])

            if work["type"] == Works.STATE_TRANSITION.value:
                result = self.transitions[
                    work["args"]["transition_id"]
                ].transition(work["args"]["state"])

            result = dict(
                id=work["args"]["id"],
                result=result
            )
            try:
                result = serialize(result)
            except Exception as e:
                print("Serialization error for result")
                print(result)
                print("from work")
                print(work)
                raise e
            self.results_queue.put(result)


import multiprocessing
import dataclasses

class Object:
    def __init__(self, *, name, init, state_mappers=[], transitions=[], sprites=[]):
        self.name = name
        self.init = init
        self.state_mappers = state_mappers
        self.transitions = transitions
        self.sprites = sprites
        self.config = None
    
    def set_config(self, config):
        self.config = config


import numpy as np
import scipy.spatial.distance
import copy

from michie.mappers.globalmapper import GlobalMapper

class ConditionNeighboursGlobalMapper(GlobalMapper):
    def __init__(self, *, condition, beacon=None):
        self.condition = condition
        self.beacon = type(self).beacon if beacon is None else beacon
    
    @staticmethod
    def beacon(state):
        beacon = state.copy()
        if "neighbours" in beacon: del beacon["neighbours"]
        return beacon

    def map(self, states, global_state):
        for id, state in enumerate(states):
            state["neighbours"] = []
            for nid, other_state in enumerate(states):
                if self.condition(states, global_state, id, nid):
                    beacon = self.beacon(other_state)
                    assert "neighbours" not in beacon, "The neighbour beacon must not have the neighbours field"
                    state["neighbours"].append(beacon)

        return states



import michie.transitions
import michie.mappers
import michie.sprites
import michie.utils
import michie.utils.init
import michie.hooks

from michie.transitions import Transition
from michie.mappers.globalmapper import GlobalMapper
from michie.mappers.statemapper import StateMapper
from michie.object import Object
from michie.sprites.sprite import Sprite
from michie.world import World
from michie.factory import factory
from michie.hooks import Hook


import numpy as np
import pygame

class HeadingSprite:
    def __init__(self, color=None, length=30, width=3):
        self.color = color
        self.length = length
        self.width = width
    
    def draw(self, *, window, state):
        color = "white"
        if "color" in state: color = state["color"]
        elif self.color is not None: color = self.color

        start = state["position"]["position"]
        end = (
            state["position"]["position"][0] + self.length*np.cos(state["position"]["heading"]),
            state["position"]["position"][1] + self.length*np.sin(state["position"]["heading"])
        )

        pygame.draw.line(window, color, start, end, self.width)


import numpy as np
import scipy.spatial.distance

from michie.mappers.globalmapper import GlobalMapper

class DistancesGlobalMapper(GlobalMapper):
    def map(self, states, global_state):
        positions = np.array([state["position"]["position"] for state in states])
        
        dists = scipy.spatial.distance.pdist(positions, "euclidean")
        dists = scipy.spatial.distance.squareform(dists)
        
        for state, dist in zip(states, dists):
            state["distances"] = dist.tolist()

        return states


from michie.mappers.globalmapper import GlobalMapper

class AddTickGlobalMapper(GlobalMapper):
    def map(self, states, global_state):
        for state in states:
            state["tick"] = global_state["tick"]

        return states





class Transition:
    @classmethod
    def requirements(cls, state):
        return True

    @classmethod
    def state_map(cls, state):
        raise NotImplementedError
    
    @classmethod
    def transition(cls, mapped_state):
        raise NotImplementedError


import pygame

class PointSprite:
    def __init__(self, radius=None, color=None):
        self.radius = radius
        self.color = color
    
    def draw(self, *, window, state):
        color = "white"
        if "color" in state: color = state["color"]
        elif self.color is not None: color = self.color

        radius = 5
        if "radius" in state: radius = state["radius"]
        elif self.radius is not None: radius = self.radius

        pygame.draw.circle(window, color, state["position"]["position"], radius)


from michie.mappers.globalmapper import GlobalMapper
from michie.mappers.statemapper import StateMapper

from michie.mappers.distances import DistancesGlobalMapper
from michie.mappers.neighbours import NeighboursGlobalMapper
from michie.mappers.communication_neighbours import CommunicationNeighboursGlobalMapper
from michie.mappers.condition_neighbours import ConditionNeighboursGlobalMapper
from michie.mappers.add_tick import AddTickGlobalMapper

from michie.mappers.range_and_bearing import RangeAndBearingStateMapper




class StateMapper:
    @classmethod
    def requirements(cls, state):
        return True

    @classmethod
    def map_state(cls, state):
        raise NotImplementedError()
    
    @classmethod
    def map_global_state(cls, state):
        raise NotImplementedError()
    
    @classmethod
    def map(cls, id, mapped_state, mapped_global_state):
        raise NotImplementedError()


import setuptools
import os

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md"), "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="michie",
    version="0.1",
    author="Federico A. Galatolo",
    author_email="federico.galatolo@ing.unipi.it",
    description="",
    url="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        "tqdm>=4.64.1",
        "pygame>=2.1.2",
        "numpy>=1.23.4",
        "schema>=0.7.5",
        "scipy>=1.9.2",
        "orjson>=3.8.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "Development Status :: 4 - Beta"
    ],
)


import numpy as np
import math
import pygame

class DroneSprite:
    def __init__(self, color=None, height=10, width=15):
        self.color = color
        self.height = height
        self.width = width
    
    def draw(self, *, window, state):
        color = "white"
        if "color" in state: color = state["color"]
        elif self.color is not None: color = self.color

        x, y = state["position"]["position"]
        angle = state["position"]["heading"]

        points = [
            (x, y - (self.height / 2)),
            (x - (self.width / 2), y + (self.height /2)),
            (x, y + (self.height / 4)),
            (x + (self.width / 2), y + (self.height / 2)),
            (x, y - (self.height / 2)),
            (x, y + (self.height / 4)),
        ]

        position = pygame.math.Vector2((x, y))
        rotated_points = [
            (pygame.math.Vector2(p) - position) \
            .rotate_rad(angle + math.pi/2) \
            for p in points
        ]

        translated_points = [(position + p) for p in rotated_points]

        pygame.draw.polygon(
            window,
            color,
            translated_points
        )



import math
import numpy as np

from michie.transitions.transition import Transition
from michie.factory import factory

class MoveTransition(Transition):
    @classmethod
    def state_map(cls, state):
        return dict(
            position=state["position"],
            speed=state["speed"]
        )
    
    @classmethod
    def transition(cls, mapped_state):
        x_k, y_k = mapped_state["position"]["position"]
        theta_k = mapped_state["position"]["heading"]
        v_k = mapped_state["speed"]["linear"]
        w_k = mapped_state["speed"]["angular"]

        if abs(w_k) <= 0.0175:
            theta_k1 = theta_k
            x_k1 = x_k + v_k*np.cos(theta_k)
            y_k1 = y_k + v_k*np.sin(theta_k)
        else:
            theta_k1 = (theta_k + w_k) % (2*math.pi)
            x_k1 = x_k + (v_k/w_k)*(np.sin(theta_k1) - np.sin(theta_k))
            y_k1 = y_k - (v_k/w_k)*(np.cos(theta_k1) - np.cos(theta_k)) 

        return dict(
            position=dict(
                position=(float(x_k1), float(y_k1)),
                heading=float(theta_k1)
            )
        )


@factory
class WrappedMoveTransitionFactory(Transition):
    @classmethod
    def state_map(cls, state):
        return MoveTransition.state_map(state)
    
    @classmethod
    def transition(cls, mapped_state):
        mapped_state = MoveTransition.transition(mapped_state)
        mapped_state["position"]["position"] = (
            mapped_state["position"]["position"][0] % cls.args["bounds"][0],
            mapped_state["position"]["position"][1] % cls.args["bounds"][1],
        )
        return mapped_state


class EulerMoveTransition(Transition):
    @classmethod
    def state_map(cls, state):
        return dict(
            position=state["position"],
            speed=state["speed"]
        )
    
    @classmethod
    def transition(cls, mapped_state):
        x_k, y_k = mapped_state["position"]["position"]
        theta_k = mapped_state["position"]["heading"]
        v_k = mapped_state["speed"]["linear"]
        w_k = mapped_state["speed"]["angular"]

        x_k1 = x_k + v_k*np.cos(theta_k)
        y_k1 = y_k + v_k*np.sin(theta_k)
        theta_k1 = (theta_k + w_k)%(2*math.pi)

        return dict(
            position=dict(
                position=(float(x_k1), float(y_k1)),
                heading=float(theta_k1)
            )
        )